// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include "Device.h"

/********************        IPs base address       *****************************/
#define SUPERNOVA_PCIE_CONTROLLER_BASE_ADDR 0xC0000000 // Pcie controller address
#define SUPERNOVA_SYSTEM_CONTROLLER 0x800000           // 0xC0800000  // system controller address
#define SUPERNOVA_SPI0_BASE_ADD 0x2A10000              // 0xD7A50000 // SPI0 controller address
#define SUPERNOVA_DMA_BASE_ADD 0xD7A53000              // DMA controller address
#define SUPERNOVA_UART0_BASE_ADD 0xD7A59000            // Uart0 controller address
#define SUPERNOVA_MAILBOX_BASE_ADD 0xC1500000          // Mailbox address
#define SUPERNOVA_TIMER_BASE_ADD 0x2A1B000             // 0xD7A5B000  // Timer controller address
#define SUPERNOVA_OTP_BASE_ADD 0xD7A5D000              // OTP controller address

#define SPI_BASE SUPERNOVA_SPI0_BASE_ADD
#define SUPERNOVA_TIMER_ADD_BASE SUPERNOVA_TIMER_BASE_ADD

// timer register
#define TIMER_LOAD(x) (SUPERNOVA_TIMER_ADD_BASE + 0x00 + (0x10 * x)) // load register
#define TIMER_CNT(x) (SUPERNOVA_TIMER_ADD_BASE + 0x04 + (0x10 * x))  // count register
#define TIMER_CTL(x) (SUPERNOVA_TIMER_ADD_BASE + 0x08 + (0x10 * x))  // control register
#define TIMER_STR(x) (SUPERNOVA_TIMER_ADD_BASE + 0x0C + (0x10 * x))  // interrupt status register read can clear interrupt

// PLL register
#define SC_PLL_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0868)
#define SC_PLL_1_CON_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0864)
#define FABRIC_PLL_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x085C)
#define FABRIC_PLL_1_CON_STUS (SUPERNOVA_SYSTEM_CONTROLLER + 0x0858)

// #define AHB_SRAM_ADDR 0x2900000      // 0xD7940000
// #define THS2_AXI_SRAM_ADDR 0x2400000 // 0xD7000000
// #define MAX_STRING_ITEM_NUM 4
// #define MAX_STRING_ITEM_LEN 12
// #define MAX_CMD_STRING_SIZE 64

#define SPI_TXREG(x) (SPI_BASE + 0x00 + 0x1000 * x)
#define SPI_RXREG(x) (SPI_BASE + 0x00 + 0x1000 * x)
#define SPI_SCLK_CON(x) (SPI_BASE + 0x04 + 0x1000 * x)
#define SPI_CTL1(x) (SPI_BASE + 0x08 + 0x1000 * x)
#define SPI_CTL2(x) (SPI_BASE + 0x0C + 0x1000 * x)
#define SPI_CTL3(x) (SPI_BASE + 0x10 + 0x1000 * x)
#define SPI_CS(x) (SPI_BASE + 0x14 + 0x1000 * x)
#define SPI_STATUS1(x) (SPI_BASE + 0x18 + 0x1000 * x)
#define SPI_STATUS2(x) (SPI_BASE + 0x1C + 0x1000 * x)
#define SPI_BATCH_CNT_RX(x) (SPI_BASE + 0x20 + 0x1000 * x)
#define SPI_BATCH_CNT_TX(x) (SPI_BASE + 0x24 + 0x1000 * x)
#define SPI_FIFO_RXTRIG_LVL(x) (SPI_BASE + 0x28 + 0x1000 * x)
#define SPI_FIFO_TXTRIG_LVL(x) (SPI_BASE + 0x2c + 0x1000 * x)
#define SPI_DUAL_QUAD_MODE(x) (SPI_BASE + 0x30 + 0x1000 * x)
#define SPI_XIP_CFG(x) (SPI_BASE + 0x34 + 0x1000 * x)
#define SPI_FLASH_ADDR_HEAD(x) (SPI_BASE + 0x38 + 0x1000 * x)

#define SCLK_DIV(divisor, prescaler) ((2L << (divisor)) * (prescaler + 1))
#define SCLK_VAL(divisor, prescaler) (((divisor) << 4L) | ((prescaler) << 0))

//SPI_CTRL1
#define LSB (0x40)
#define MSB (0x00)
#define CPHA_H (0x20)
#define CPHA_L (0x00)
#define CPOL_H (0x10)
#define CPOL_L (0x00)
#define MST_MODE (0x04)
#define SLV_MODE (0x00)

#define TX_ENABLE (0x08)
#define TX_DISABLE (0x00)
#define RX_ENABLE (0x80)
#define RX_DISABLE (0x00)

#define FIFO_WIDTH32 (0x02)
#define FIFO_WIDTH16 (0x01)
#define FIFO_WIDTH8 (0x00)
#define FIFO_DEPTH (16)
#define BATCH_MAX_SIZE (0x10000)

#define SPI_FIFO_DEPTH (FIFO_DEPTH)
#define SPI_BACKUP_OFFSET (0x10000) //64K
#define SPI_CS0 (0x01)

// Mode  CPOL    CPHA
// 0     0       0
// 1     0       1
// 2     1       0
// 3     1       1

#define SPI_MODE0 (CPOL_L | CPHA_L)
#define SPI_MODE1 (CPOL_L | CPHA_H)
#define SPI_MODE2 (CPOL_H | CPHA_L)
#define SPI_MODE3 (CPOL_H | CPHA_H)

//SPI_CTRL1
#define LSB (0x40)
#define MSB (0x00)
#define CPHA_H (0x20)
#define CPHA_L (0x00)
#define CPOL_H (0x10)
#define CPOL_L (0x00)
#define MST_MODE (0x04)
#define SLV_MODE (0x00)

#define TX_ENABLE (0x08)
#define TX_DISABLE (0x00)
#define RX_ENABLE (0x80)
#define RX_DISABLE (0x00)

#define FIFO_WIDTH32 (0x02)
#define FIFO_WIDTH16 (0x01)
#define FIFO_WIDTH8 (0x00)
#define FIFO_DEPTH (16)
#define BATCH_MAX_SIZE (0x10000)

#define SPI_FIFO_DEPTH (FIFO_DEPTH)
#define SPI_BACKUP_OFFSET (0x10000) //64K
#define SPI_CS0 (0x01)

// SPI_CTRL2
#define BATCH_TINT_EN (0x1L << 7)
#define BATCH_RINT_EN (0x1L << 6)
#define MODDEF (0x1L << 3)
#define ERR_INT_EN (0x1l << 2)
#define TX_INT_EN (0x1L << 1)
#define RX_INT_EN (0x1L << 0)

// SPI_CTRL3
#define BATCH_DISABLE (0x1L << 4)
#define BATCH_ENABLE (0x0L << 4)
#define FIFO_RESET (0x03)
#define SW_RST_TX (0x02)
#define SW_RST_RX (0x01)
#define RX_DMA_EN (0x04)
#define TX_DMA_EN (0x08)

// STATUS1
#define BATCH_TXDONE (0x1L << 5)
#define BATCH_RXDONE (0x1L << 4)
#define MODF_CLR (0x1L << 3)
#define TX_ALMOST_EMPTY (0x1L << 1)
#define RX_ALMOST_FULL (0x1L << 0)

// STATUS2
#define SSN_IN (0x1L << 7)
#define TXFIFO_FULL (0x1L << 5)
#define TXFIFO_EMPTY (0x1L << 1)
#define RXFIFO_EMPTY (0x1L << 4)
#define RXFIFO_FULL (0x1L << 0)

// SPI_DUAL_QUAD_MODE
// #define SYS_QPI_DPI_EN
#define HOLD_OUTPUT (0x80)
#define HOLD_OE (0x40)
#define HOLD_CTL (0x20)
#define WP_OUTPUT (0x10)
#define WP_OE (0x08)
#define WP_CTL (0x04)
#define QPI_ENABLE (0x02)
#define DPI_ENABLE (0x01)
#define SPI_XFER_TX 0
#define SPI_XFER_RX 1
#define SPI_XFER_TXRX 2

// SPI_XIP_CFG
// #define SYS_XIP_EN
#define XIP_EN (0x1L << 0)
#define FLASH_SW_CFG (0x0L << 1)
#define FLASH_HW_CFG (0x1L << 1)
#define FLASH_CTNU_MODE (0x1 << 4)
#define DUMMY_CYCLE_0 (0x0 << 5)
#define DUMMY_CYCLE_1 (0x1 << 5)
#define DUMMY_CYCLE_2 (0x2 << 5)
#define DUMMY_CYCLE_3 (0x3 << 5)
#define DUMMY_CYCLE_4 (0x4 << 5)
#define DUMMY_CYCLE_5 (0x5 << 5)
#define DUMMY_CYCLE_6 (0x6 << 5)
#define DUMMY_CYCLE_7 (0x7 << 5)

#define XIP_FLASH_ADDR_OFFSET 0x100
// batch mode or no batch mode
#define SPI_BATCH_MODE 0
#define SPI_NO_BATCH_MODE 1

// transfer mode
#define SPI_POLL_MODE 0
#define SPI_INT_MODE 1
#define SPI_DMA_MODE 2

// wire mode
#define SPI_NORMAL_MODE 1
#define SPI_DPI_MODE 2
#define SPI_QPI_MODE 4

#define MIN(A, B) ((A) > (B)) ? (B) : (A)
#define MAX(A, B) ((A) > (B)) ? (A) : (B)
#define SPI_BAUDRATE (45 * 1000000) //actual value is 50MHz
#define SPI_AHB_CLK (500 * 1000000) //500MHz

#define TIMEOUT_1S 1000000 // CommonDelayUsFunc(TIMEOUT_1S)
#define TIMEOUT_2uS 2      // CommonDelayUsFunc(TIMEOUT_2uS)
#define TIMEOUT_20mS 2000  // CommonDelayUsFunc(TIMEOUT_20mS)

//Flash command
#define FLS_CMD_RDID (0x9F)
#define Flash_CMD_REMS (0x90)
#define Flash_MID_ADDR (0x000000)
#define Flash_DID_ADDR (0x000001)
#define FLS_CMD_RSRL (0x05)
#define FLS_CMD_RSRH (0x35)

#define FLS_CMD_READ (0x03)
#define FLS_CMD_FastREAD (0x0b)
#define FLS_CMD_WREN (0x06)
#define FLS_CMD_WRDIS (0x04)
#define FLS_CMD_SE (0x20)
#define FLS_CMD_BE_32K (0x52) // 32K block
#define FLS_CMD_BE_64K (0xD8) // 64K block
#define FLS_CMD_CE (0xC7)     // (0x60)
#define FLS_CMD_PP (0x02)
#define FLS_CMD_FSRD (0xeb)
#define FLS_CMD_WRSR (0x01)

// Flash Status
#define FLS_STS_QE (0x01 << 9)
#define FLS_STS_SRP1 (0x01 << 8)
#define FLS_STS_SRP0 (0x01 << 7)
#define FLS_STS_BP4 (0x01 << 6)
#define FLS_STS_BP3 (0x01 << 5)
#define FLS_STS_BP2 (0x01 << 4)
#define FLS_STS_BP1 (0x01 << 3)
#define FLS_STS_BP0 (0x01 << 2)
#define FLS_STS_WEL (0x01 << 1)
#define FLS_STS_WIP (0x01 << 0)

#define FLS_BLOCK_SIZE (0x00010000) //64K
#define FLS_BASE_BLOCK(n) (n * FLS_BLOCK_SIZE)
#define FLS_SECTOR_SIZE (0x00001000) // 4k
#define FLS_BASE_SECTOR(n) (n * FLS_SECTOR_SIZE)
#define FLS_PAGE_SIZE (0x00000100) //256byte
#define FLS_BASE_PAGE(n) (n * FLS_PAGE_SIZE)
#define FLS_MAX_WRITE_SIZE (FLS_PAGE_SIZE)

typedef unsigned char            U8;
typedef unsigned char*           P8;
typedef volatile unsigned char*  VP8;
typedef unsigned short           U16;
typedef unsigned short*          P16;
typedef volatile unsigned short* VP16;
typedef unsigned int             U32;
typedef unsigned int*            P32;
typedef volatile unsigned int*   VP32;
typedef signed int               S32;
typedef signed char              S8;
typedef unsigned long long       U64;
typedef unsigned int             FlashAddr_T;

typedef struct _SPI_InitTypeDef_ {
  U32 ahbBusFreq;
  U32 sclk;
  U8  mstSlvMode;
  U8  workMode;
  U8  fifoWidth;
  U8  firstBit;
  U8  rxTrigLevel;
  U8  txTrigLevel;
  U8  intCfg;
  U8  batchMode;
  U8  instance;
  U32 timeout; //check status time limit
  U8  cs;
  U8  inited;
} SPI_InitTypeDef;

typedef enum {
  SPI_OK      = 0x00, //SPI init or config or satrt transfer OK
  SPI_DONE    = 0x01, //SPI transfer success
  SPI_ERROR   = 0x02, //SPI config or transfer error
  SPI_IDLE    = 0x03, //SPI init or config or satrt transfer OK
  SPI_BUSY    = 0x04, //SPI is busy, it is running at current time
  SPI_TIMEOUT = 0x05, //SPI transfer timeout error
} SPI_StatusTypeDef;

// #define read_mreg8(addr) (*(reinterpret_cast<VP8>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))))
// #define write_mreg8(addr, val) (*(reinterpret_cast<VP8>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))) = (U32) (val))
// #define read_mreg32(addr) (*(reinterpret_cast<VP32>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))))
// #define write_mreg32(addr, val) (*(reinterpret_cast<VP32>((reinterpret_cast<volatile uint8_t*>(m_bar2Vaddr) + addr))) = (U32) (val))

typedef enum TimerMode {
  TIMER_MODE_FREERUN = 0x00, //count goes from 65535 to 0 and loops back
  TIMER_MODE_CYCLIC  = 0x02, //count goes from initial count value to 0 and loops back
  TIMER_MODE_SINGLE  = 0x03, //count goes form initial count value to 0 and stop
} TimerMode_ENUM;

typedef enum { FALSE = 0, TRUE = 1 } BOOL;

class Flash : public jsf_int::NodeLeaf {
  static constexpr uint32_t SPI_FLASH_BUFFERSIZE = 2 * 1024 * 1024; // SPI Flash Size

  static constexpr uint32_t ESecure_A_ADDR = 0x1000;
  static constexpr uint32_t ESecure_B_ADDR = 0x21000;
  static constexpr uint32_t ESecure_SIZE   = 0x20000;

  static constexpr uint32_t ZSP_A_ADDR = 0x41000; // 0x1F000 , 0x13B79 (Actual) // Max Size: 124K = 126976 bytes
  static constexpr uint32_t ZSP_B_ADDR = 0x61000;
  static constexpr uint32_t ZSP_SIZE   = 0x1F000;

  static constexpr uint32_t PcieCtlPatch_ADDR = 0x81000;
  static constexpr uint32_t PcieCtlPatch_SIZE = 0x20000;

  static constexpr uint32_t PciePhyPatch_A_ADDR = 0xA1000;
  static constexpr uint32_t PciePhyPatch_B_ADDR = 0xC1000;
  static constexpr uint32_t PciePhyPatch_SIZE   = 0x20000;

  static constexpr uint32_t PcieFw_ADDR = 0xE1000;
  static constexpr uint32_t PcieFw_SIZE = 0x11800;

  static constexpr uint32_t SPI_FLASH_ADDR = 0x0;
  static constexpr uint32_t SPI_FLASH_SIZE = 0x200000;

  static constexpr uint32_t SC_ADDR = 0x100000;
  static constexpr uint32_t SC_SIZE = 0x80000;

  static constexpr uint32_t Reserved_ADDR = 0xF2800;
  static constexpr uint32_t Reserved_SIZE = 0x10D800;

private:
  Device& m_device;

public:
  Flash(Device& Device, const std::string& Name) : NodeLeaf(Name), m_device(Device) {
  }

  virtual json Invoke(const json& Parameters = {}, bool Iteration = false) override;

private:
  void SPIMasterInitFunc(U8 SPINum, U32 IpClk);
  void SPI_Init(SPI_InitTypeDef* SPI_InitType);

  void ReadFlash(U32 Addr, U32 Len, const std::string& Read_Dump_Flash_Mem_File);
  void SPIFlashReadFunc(U32 Addr, U8* pData, U32 Count);
  void ReadVerifyFlash(U32 Addr, U32 Len, U8* pBuff);

  void DumpFlashBuffer(U8* fbuff, U32 len, std::string Read_Flash_File);

  void              SPIChipSelectFunc(void);
  void              SPIMasterTransmitFunc(U8* Buff, U32 TxLen);
  SPI_StatusTypeDef SPI_Transmit(SPI_InitTypeDef* spi_instance, U8* pTxData, U32 Len);
  void              SPIMasterReceiveFunc(U8* Buff, U32 RxLen);
  SPI_StatusTypeDef SPI_Receive(SPI_InitTypeDef* spi_instance, U8* pRxData, U32 Len);
  void              SPIChipDeselectFunc(void);
  SPI_StatusTypeDef SPI_CheckStatus(U32 Reg, U32 Flag, U8 Level, U32 TimeOut);
  void              CommonDelayUsFunc(U32 Count);
  void              Usleep(U32 Us);
  U32               TimerReadCountFunc(U32 TimerIx);

  int  VerifyFlashedRegion(U32 mem_addr, U32 mem_len, std::string flash_file);
  void WriteFlash(const char* Fname, U32 Addr, U32 Len);
  void SPIFlashPageProgramFunc(U8* src_buff, U32 dest_flash_addr, U32 len);
  void EraseFlash(U32 Addr, U32 Len);
  void SPIFlashEraseFunc(U32 flash_addr, U32 len);
  void SPIFlashSectorEraseFunc(U32 addr);
  void SPIFlash_Disable_Protect(void);
  int  SPIFlash_Checkbusy(void);
  void SPIFlash_Write_Disable(void);
  void SPIFlash_Write_Enable(void);
};
