// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <fcntl.h>
#include <Reset.h>
#include <Logging.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#define IOCTL_ALL_IP_SOFT_RESET _IO('k', 0xc0)
json Reset::Invoke(const json& Parameters, bool Iteration) {
  if (Iteration) {
    return {};
  }
  const auto bdf = m_device.GetBdf();

  if (Parameters.contains("type")) {
    if (Parameters["type"] == "flr") { // flr (sudo ./testapps/jmamgmt --live devices/0000:02:00.0/reset '"type":"flr"')

      if (m_device.GetPciDevice().device_id == 0x5070) { // PF device
        const auto    sriov_pf_path = "/sys/bus/pci/devices/" + bdf + "/sriov_numvfs";
        std::ifstream sriov_pf_file(sriov_pf_path);
        int           val_sriov_numvfs = -1;
        sriov_pf_file >> val_sriov_numvfs;
        if (val_sriov_numvfs != 0) {
          throw std::runtime_error("Please destroy all VF devices corresponding to PF device " + bdf);
        }

        const auto device_path = "/dev/ama_transcoder" + std::to_string(m_device.GetDeviceNum());
        auto       dev         = open(device_path.c_str(), O_RDWR);
        if (dev < 0) {
          printf("Device open failed!\n");
          exit(1);
        }

        int ret = ioctl(dev, IOCTL_ALL_IP_SOFT_RESET);
        if (ret < 0) {
          logFatal("IOCTL_ALL_IP_SOFT_RESET: failed");
        }
        return 1;
      } else if (m_device.GetPciDevice().device_id == 0x5071) {                    // VF device
        auto        sysfs_driver_path = "/sys/bus/pci/devices/" + bdf + "/driver"; // VM powered-off/unbound
        struct stat buffer;
        if (stat(sysfs_driver_path.c_str(), &buffer) == 0) {
          throw std::runtime_error("Shutdown VM and apply reset");
        }
        // Reset
        const auto    sysfs_reset = "/sys/bus/pci/devices/" + bdf + "/reset"; // echo 1 > /sys/bus/pci/devices/0000\:01\:00.1/reset
        unsigned char val_reset   = 1;
        auto          fd_reset    = open(sysfs_reset.c_str(), O_WRONLY);
        if (fd_reset < 0) {
          logFatal("reset: file Not opened");
          m_device.fail();
        }
        auto ret = write(fd_reset, std::to_string(val_reset).c_str(), 1);
        if (ret != 1) {
          logFatal("reset: write failed");
          close(fd_reset);
          m_device.fail();
        }
        close(fd_reset);
        return 1;
      } else {
        throw std::runtime_error("invalid device_id");
      }
    }
    throw std::runtime_error("invalid reset type");
  }
  if (!Parameters.is_null()) {
    throw std::runtime_error("reset type required / invalid type");
  }
  return {};
}
